/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

'use strict';

var path = require('path');
var fs = require('fs');
var os = require('os');

var evt = require("./evt");
var env = require("./env");
var log = require("./log");
var md5 = require("./md5");
var agt = require("./agt");
var ver = require(ISEDORA_PKG);
var msg = require("./msg");
var rti = require("./rti");
var jsl = require("./jsl");

var OSTYPE = {
  UnknownOS: 0,
  Linux: 0x2000,
  Android: 0x3000,
  iOS: 0x8000,
  MacOSX_10_4:  0x1004,
  MacOSX_10_5:  0x1005,
  MacOSX_10_6:  0x1006,
  MacOSX_10_7:  0x1007,
  MacOSX_10_8:  0x1008,
  MacOSX_10_9:  0x1009,
  MacOSX_10_10: 0x100A,
  Win2000:  0x4105,
  WinXP:    0x4106,
  WinVista: 0x4107,
  Windows7: 0x4108,
  Windows8: 0x4109,
  Windows:  0x4000
};

exports.OSTYPE = OSTYPE;

var ID = {
  uuid: 'uuid',
  osname: 'osname',
  ostype: 'ostype',
  osbits: 'osbits',
  mybits: 'mybits',
  displayLang: 'displayLanguage',
  logonName: 'logonName',
  userFullName: 'userFullName',
  userIsRoot: 'userIsRoot',
  folders: 'folders',
  appData: 'appData',
  cache: 'cache',
  logs: 'logs',
  sharedPublic: 'sharedPublic',
  applications: 'applications',
  temp: 'temp',
  image: 'image',
  audio: 'audio',
  video: 'video',
  ifaces: 'ifaces',
  thumbColor: 'thumbColor',
  imageRoots: 'imageRoots',
  audioRoots: 'audioRoots',
  videoRoots: 'videoRoots',
  topOrder: 'topOrder',
  topGroup: 'topGroup',
  dirFirst: 'dirFirst',
  hideExt: 'hideExt',
  enableImage: 'enableImage',
  enableAudio: 'enableAudio',
  enableVideo: 'enableVideo',
  modOff: 'modOff',
  childCount: 'childCount',
  hideEmpty: 'hideEmpty',
  plugins: 'plugins',
  plugoff: 'plugoff'
};

var PATHS = {
  cache: 'Cache',
  minfo: 'Minfo',
  thumb: 'Thumb',
  scale: 'Scale',
  icons: 'Icons'
};

var FILE_OPTS = {encoding: 'utf8'};

var DATA_PATH   = ver['dataPath'];
var CONFIG_NAME = ver.name + '.cfg';
var LOG_NAME    = ver.name + '.log';
var PID_NAME    = ver.name + '.' + process.pid;
var PUB_DIR     = ISEDORA_DIR + '/pub';

var REST_ROUTE = "/rest";
var DLNA_ROUTE = "/dlna";
var FACE_ROUTE = "/face";

//-----------------------------------------------------------------------------
// Defauls
//-----------------------------------------------------------------------------

var DEFAULTS = {
  startup: 0,
  power: 2,
  httpPort: 8010,
  thumbColor: '000000',
  agentCount: 4,
  logLevel: log.LEVELS.info,
  logParts: 0,
  nicBlock: {},
  ipBlock: {},
  aliveInterval: 60,
  searchInterval: 30,
  checkStamp: 60,
  udpTTL: 4,
  devHide: {
    "urn:schemas-upnp-org:device:InternetGatewayDevice:1": true,
    "urn:schemas-upnp-org:device:MediaServer:1": true
  },
  devcfg: {},
  serverName: ver.description + ' ' + parseInt(ver.version, 10),
  plugins: {},
  plugoff: {},
  topOrder: [rti.mtype.video, rti.mtype.audio, rti.mtype.image],
  topGroup: false,
  dirFirst: true,
  hideExt: true,
  childCount: true,
  hideEmpty: true,
  enableImage: true,
  enableAudio: true,
  enableVideo: true,
  imageRoots: null,
  audioRoots: null,
  videoRoots: null,
  imageScale: '3840×2160',
  imageRotate: true
};

var ready = false;
var sysinfo = {};
var config = {};
var appData = null;
var logsDir = null;
var cacheDir = null;
var tempDir = null;
var folders = null;

var logFilePath = null;
var pidFilePath = null;

var thumbCache;
var minfoCache;
var scaleCache;
var iconsCache;
var isWinXP = false;

function loadConfig() {
  var cfile = path.join(appData, CONFIG_NAME);
  if (fs.existsSync(cfile)) {
    var data = fs.readFileSync(cfile, FILE_OPTS);
    try {
      config = JSON.parse(data);
    }
    catch (e) {
      log.error("Error parsing config file: %s", e);
      fs.unlink(cfile);
      config = {};
    }
  }
}

function saveConfig(obj, postEvent, cb) {
  var cfile = path.join(appData, CONFIG_NAME);
  try {
    fs.writeFile(cfile, JSON.stringify(obj, null, 2), FILE_OPTS, function (err) {
      if (err) {
        if (cb) cb(err);
      }
      else {
        config = obj;
        log.setLevel(config.logLevel);
        log.setParts(config.logParts);
        if (postEvent) {
          setImmediate(evt.postConfigChange());
        }
        if (cb) cb();
      }
    });
  }
  catch (e) {
    if (cb) cb(e);
  }
}

function createDir(path, mode) {
  try {
    if (!fs.existsSync(path)) {
      log.trace('Creating dir: %s', path);
      fs.mkdirSync(path);
      if (mode) {
        fs.chmodSync(path, mode);
      }
    }
  } catch (e) {
    throw 'Error creating folder "' + path + '"';
  }
}

function prepareFolders() {

  folders = sysinfo[ID.folders];

  if (!folders) throw "Sysinfo folders not specified";

  var ix = userIsRoot() ? 1 : 0;
  var mode = (userIsRoot() && env.isWin()) ? 0x1ff : 0;

  // App Data

  appData = folders[ID.appData][ix];

  if (!appData) throw "AppData folder not specified";

  appData = path.join(appData, DATA_PATH);

  createDir(appData, mode);

  // PID File

  var pidDir = path.join(env.ifMac("/var/tmp", folders[ID.appData][1]), DATA_PATH);

  createDir(pidDir, 0x1ff);

  pidFilePath = path.join(pidDir, PID_NAME);

  // Cache

  cacheDir = folders[ID.cache][ix];

  if (cacheDir.length === 0) {
    cacheDir = path.join(appData, PATHS.cache);
  }
  else {
    cacheDir = path.join(cacheDir, DATA_PATH);
  }

  createDir(cacheDir);

  thumbCache = path.join(cacheDir, PATHS.thumb);
  minfoCache = path.join(cacheDir, PATHS.minfo);
  scaleCache = path.join(cacheDir, PATHS.scale);
  iconsCache = path.join(cacheDir, PATHS.icons);

  createDir(thumbCache);
  createDir(minfoCache);
  createDir(scaleCache);
  createDir(iconsCache);

  // Logs

  logsDir = folders[ID.logs][ix];

  if (logsDir.length === 0) {
    logsDir = appData;
  }
  else {
    logsDir = path.join(logsDir, DATA_PATH);
  }

  createDir(logsDir);

  logFilePath = path.join(logsDir, LOG_NAME);

  // Temp Dir

  tempDir = path.join(os.tmpdir(), DATA_PATH);

  createDir(tempDir);

  // Load Config

  loadConfig();

  jsl.mergeObject(DEFAULTS, config);

  defaultMediaFolders(ID.imageRoots, folders[ID.image]);
  defaultMediaFolders(ID.audioRoots, folders[ID.audio]);
  defaultMediaFolders(ID.videoRoots, folders[ID.video]);

  log.setLevel(config.logLevel);

  log.setParts(config.logParts);

  log.debug("Log file: %s", logFilePath);

  log.initLogFile(logFilePath);

  msg.load(sysinfo[ID.displayLang]);

  ready = true;

  log.debug("Config ready");

  evt.postConfigReady();

}

function defaultMediaFolders(name, folders) {

  if (config[name]) return;

  var obj = {};

  for (var i = 0; i < folders.length; i++) {
    var f = folders[i];
    obj[f.path] = {name: f.name, sort: 0, uid: md5(f.path)};
  }

  config[name] = obj;

}

function on_sysinfo(err, res) {

  if (err) {
    evt.throwGlobal(err);
  }

  try {

    sysinfo = res;

    var uuid = sysinfo[ID.uuid];

    if (!uuid) {
      //noinspection ExceptionCaughtLocallyJS
      throw "Unable to get system unique id.";
    }

    sysinfo[ID.uuid] = md5.uuid(uuid);

    var ostype = sysinfo[ID.ostype];

    isWinXP = ((ostype & OSTYPE.Windows) && (ostype < OSTYPE.WinVista)) ? true : false;

    prepareFolders();

  } catch (ex) {
    evt.throwGlobal(ex);
  }

}

function uuid() {
  return sysinfo[ID.uuid];
}

function userIsRoot() {
  return sysinfo[ID.userIsRoot];
}

// Exports ====================================================================

exports.getConfig = function() { return config; };

exports.getPluginConfig = function(name) { return config.plugins[name] };
exports.setPluginConfig = function(name, value) { config.plugins[name] = value };
exports.pluginEnabled   = function(name) {
  var status = config.plugoff[name];
  return (status !== true);
};


exports.saveConfig = saveConfig;

exports.saveSelf = function(postEvent, cb) {
  saveConfig(config, postEvent, cb);
};

exports.sysinfo = function() { return sysinfo };

exports.ifaces = function() { return sysinfo[ID.ifaces] };

exports.uuid = uuid;

exports.userIsRoot = userIsRoot;

exports.appData = function () { return appData };

exports.cacheDir = function () { return cacheDir };

exports.logsDir = function () { return logsDir };

exports.ready = function () { return ready };

exports.osname = function () { return sysinfo[ID.osname] };

exports.ostype = function () { return sysinfo[ID.ostype] };

exports.osbits = function () { return sysinfo[ID.osbits] };

exports.mybits = function () { return sysinfo[ID.mybits] };

exports.lang = function () { return sysinfo[ID.displayLang] };

exports.userFullName = function () { return sysinfo[ID.userFullName] };

exports.thumbCache = function () { return thumbCache };

exports.minfoCache = function () { return minfoCache };

exports.scaleCache = function () { return scaleCache };

//exports.THUMB_DLNA = function() { return 160 };

exports.httpPort = function () { return config.httpPort };

exports.setHttpPort = function (val) { config.httpPort = val };

exports.serverName = function () { return config.serverName };

exports.isWinXP = function () { return isWinXP };

exports.pubDir = function () { return PUB_DIR };

exports.tempDir = function () { return tempDir };

exports.logFilePath = function () { return logFilePath };

exports.pidFilePath = function () { return pidFilePath };

exports.logonName = function () { return sysinfo[ID.logonName] };

exports.restPath = function (path) {
  return (path) ? REST_ROUTE + path : REST_ROUTE;
};
exports.dlnaPath = function (path) {
  return (path) ? DLNA_ROUTE + path : DLNA_ROUTE;
};
exports.facePath = function (path) {
  return (path) ? FACE_ROUTE + path : FACE_ROUTE;
};

exports.rest = function(app) {
  return app.get(REST_ROUTE);
};

exports.dlna = function(app) {
  return app.get(DLNA_ROUTE);
};

exports.face = function(app) {
  return app.get(FACE_ROUTE);
};

exports.agentCount = function() {
  return config.agentCount;
};

exports.thumbColor = function() {
  var clr = parseInt(config[ID.thumbColor], 16);
  return (isNaN(clr) ? 0 : clr);
};

exports.nicBlocked = function(name) {
  return (config.nicBlock[name] === true);
};

exports.ipBlocked = function(addr) {
  return (config.ipBlock[addr] === true);
};

exports.addBlockedIP = function(addr) {
  config.ipBlock[addr] = true;
};

exports.isSysNetFace = function(name) {
  return (sysinfo[ID.ifaces][name] !== undefined);
};

var serverHeader = ver.name + '/2.0 UPnP/1.0 DLNADOC/1.50';
//var serverHeader = 'MacOS, UPnP/1.0, iSedora/2.0.0';

exports.serverHeader = function() {
  return serverHeader;
};

exports.searchInterval = function() {
  return config.searchInterval;
};

exports.aliveInerval = function() {
  return config.aliveInterval;
};

exports.udpTTL = function() {
  return config.udpTTL;
};

exports.devHidden = function(uuid) {
  return (uuid && config.devHide[uuid] === true);
};

exports.setDevHidden = function(uuid, hidden) {
  config.devHide[uuid] = hidden;
};

exports.imageRoots = function() {
  return config[ID.imageRoots];
};

exports.audioRoots = function() {
  return config[ID.audioRoots];
};

exports.videoRoots = function() {
  return config[ID.videoRoots];
};

exports.topOrder = function() {
  return config[ID.topOrder];
};

exports.topGroup = function() {
  return config[ID.topGroup];
};

exports.dirFirst = function() {
  return config[ID.dirFirst];
};

exports.hideEmpty = function() {
  return config[ID.hideEmpty];
};

exports.hideExt = function() {
  return config[ID.hideExt];
};

exports.enableImage = function() {
  return config[ID.enableImage];
};

exports.enableAudio = function() {
  return false; // config[ID.enableAudio];
};

exports.enableVideo = function() {
  return false; // config[ID.enableVideo];
};


exports.childCount = function() {
  return config[ID.childCount];
};

exports.getSysImageFolders = function() {
  return (folders) ? folders[ID.image] : null;
};

exports.getSysAudioFolders = function() {
  return (folders) ? folders[ID.audio] : null;
};

exports.getSysVideoFolders = function() {
  return (folders) ? folders[ID.video] : null;
};

exports.imageScale  = function() { return config.imageScale };
exports.imageRotate = function() { return config.imageRotate };

exports.getDevCfg = function(uuid) { return config.devcfg[uuid] };
exports.setDevCfg = function(uuid, dcfg) {
  config.devcfg[uuid] = dcfg;
  saveConfig(config, false);
};

exports.getCheckStamp = function() {
  return (config.checkStamp * 1000);
};

exports.getPower = function() {
  return config.power;
};

// Events Listener ============================================================

evt.onAgentReady(function () {
  agt.sysinfo(on_sysinfo);
});

