/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

'use strict';

var events = require('events');
var spawn  = require('child_process').spawn;

var byline = require('byline');
var log = require('../log');
var env = require('../env');

var ENCODING = env.ifMac("utf8", "utf8");
var ENDLINE  = env.ifMac("\n", "\n");
var READYRES = "ready";
var BYLINE_OPTS = {keepEmptyLines: false};

var AGENT_COUNTER = 0;

//-----------------------------------------------------------------------------

function thunkcb(err) {
  if (err) {
    log.error('%s', err);
  }
}

//-----------------------------------------------------------------------------

function Agent(agentPath, agentArgs) {

  var self = this;

  var emitter    = new events.EventEmitter();
  var exitCalled = false;
  var shouldStop = false;
  var running  = false;
  var ready    = false;
  var cbary    = [];
  var name     = "iSedoraAgent " + AGENT_COUNTER++;
  var cmdcount = 0;
  var agent;
  var stdout;
  var stderr;

  function onStdOut(line) {

    if (line === READYRES) {
      ready = true;
      emitReady();
      checkStop();
      return;
    }

    var cb = cbary.shift();

    if (!cb) {
      log.error('Undefined agent callback!');
      checkStop();
      return;
    }

    var obj;

    try {
      obj = JSON.parse(line);
    } catch (ex) {
      cb(ex);
      checkStop();
      return;
    }

    var err = obj.err;
    var res = obj.res;

    if (err !== undefined) {
      cb(err);
    }
    else if (res !== undefined) {
      cb(null, res);
    }
    else {
      log.error('Agent invalid response: %s', line);
      cb(new Error('Invalid response'));
    }

    checkStop();

  }

  function checkStop() {
    if (shouldStop && cbary.length === 0) {
      stop();
    }
  }

  function onStdErr(line) {
    log.error('[AGENT] %s', line);
  }

  function emitError(err) {
    emitter.emit('error', err);
  }

  function emitReady() {
    emitter.emit('ready', self);
  }

  function emitExit() {
    emitter.emit('exit', self);
  }

  function onExit(code, signal) {
    if (exitCalled) return;
    exitCalled = true;
    running = false;
    ready   = false;
    agent   = null;
    stdout  = null;
    stderr  = null;
    while (cbary.length > 0) {
      var cb = cbary.shift();
      cb(new Error('Agent terminated'));
    }
    var msg = 'Agent exited';
    if (typeof code === 'number')   msg += ' (exit code: ' + code + ')';
    if (typeof signal === 'string') msg += ' (signal: ' + signal + ')';
    if (shouldStop) {
      log.debug(msg);
    }
    else {
      log.error(msg);
    }
    emitExit();
  }

  function onError(err) {
    log.error('[AGENT ERROR] %s', err);
    emitError(err);
    onExit();
  }

  function start() {
    if (running) return false;
    exitCalled = false;
    shouldStop = false;
    cmdcount = 0;
    agent = spawn(agentPath, agentArgs, { cwd: process.cwd(), env: process.env, detached: true});
    agent.stdout.setEncoding(ENCODING);
    agent.stderr.setEncoding(ENCODING);
    stdout = byline.createStream(agent.stdout, BYLINE_OPTS);
    stderr = byline.createStream(agent.stderr, BYLINE_OPTS);
    stdout.on('data', onStdOut);
    stderr.on('data', onStdErr);
    agent.on('exit', onExit);
    agent.on('error', onError);
    running = true;
    return true;
  }

  function stop() {
    if (running) {
      try {
        shouldStop = true;
        agent.stdin.end();
        agent.kill('SIGINT');
      } catch (ex) {
        log.error('Agent stop(): %s', ex);
      }
    }
  }

  function interrupt() {
    if (running) {
      if (shouldStop) return;
      shouldStop = true;
      log.debug('Agent interrupted');
      if (cbary.length === 0) {
        stop();
      }
    }
  }

  function send(data, cb) {
    if (typeof cb !== 'function') cb = thunkcb;
    if (!ready) {
      return cb(new Error('Agent not ready'))
    }
    try {
      var body = JSON.stringify(data);
      agent.stdin.write(body + ENDLINE, ENCODING);
      cbary.push(cb);
      cmdcount++;
    } catch (ex) {
      cb(ex);
    }
  }

  // Public Interface

  self.on    = function(event, listener) {emitter.on(event, listener)};
  self.once  = function(event, listener) {emitter.once(event, listener)};
  self.removeAllListeners = function() {emitter.removeAllListeners()};

  self.start = start;
  self.stop  = stop;
  self.send  = send;
  self.interrupt = interrupt;

  self.ready    = function() {return ready};
  self.cbcount  = function() {return cbary.length};
  self.running  = function() {return running};
  self.cmdcount = function() {return cmdcount};

  self.toString = function() {return name};

}

module.exports = Agent;


