/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

'use strict';

var fs    = require('fs');
var np    = require('path');
var async = require('async');

var log   = require('../log');
var fmt   = require('../fmt');
var fop   = require('../fop');
var jsl   = require('../jsl');
var kvdb  = require('./kvdb');

var SlotMap = require('./smap').SlotMap;

var FOPTS = {encoding: 'utf8'};
var DEF_INTERVAL = 3000;

var PARALLEL_LIMIT = 4;

function noop() {}

//-----------------------------------------------------------------------------
// Size Cache
//-----------------------------------------------------------------------------

function SizeCache(cacheName, ext, minlen) {

  var self = this;
  var slot = new SlotMap();
  var path = null;
  var elen = ext.length;

  function load(cachePath) {
    var start = Date.now();
    fs.readdir(cachePath, function(err, fnames) {
      if (err) {
        log.error('Cache %s error reading dir: %s', cacheName, cachePath);
        return;
      }
      path = cachePath;
      var flen = fnames.length;
      var count = 0;
      for (var i = 0; i < flen; i++) {
        var name = fnames[i];
        var nlen = name.length;
        if (name.charAt(0) !== '.' && nlen >= minlen && name.substring(nlen - elen) === ext) {
          var key = name.substring(0, nlen - elen);
          slot.set(key, -1);
          count++;
        }
      }
      log.debug('Cache %s loaded %d items in: %s sec.', cacheName, count, fmt.millisPeriod(start));
    });
  }

  function stat(cb) {

    var count = 0;
    var size  = 0;
    var refresh = [];

    slot.each(function(key, val) {
      count++;
      if (val > -1) {
        size += val;
      }
      else {
        refresh.push(key);
      }
    });

    if (refresh.length === 0) {
      return cb(null, {count: count, size: size});
    }

    async.eachLimit(refresh, PARALLEL_LIMIT,
      function(key, acb) {
        fop.fileSize(filePath(key), function(err, fsize) {
          if (!err) {
            slot.set(key, fsize);
            size += fsize;
          }
          acb();
        });
      },
      function() {
        cb(null, {count: count, size: size});
      }
    );

  }

  function erase(cb) {
    var keys = slot.allkeys();
    slot.clear();
    async.eachLimit(keys, PARALLEL_LIMIT,
      function(key, acb) {
        fs.unlink(filePath(key), function() { acb() });
      },
      cb
    );
  }

  function filePath(key) {
    return np.join(path,  key + ext);
  }

  function get(key) {
    return slot.get(key);
  }

  function set(key, val) {
    slot.set(key, val);
  }

  function has(key) {
    return slot.has(key);
  }

  function del(key) {
    slot.del(key);
    if (path) {
      fs.unlink(filePath(key), noop);
    }
  }

  // Public Interface

  self.get = get;
  self.set = set;
  self.has = has;
  self.del = del;

  self.load  = load;
  self.stat  = stat;
  self.erase = erase;

  self.path = function() { return path };

  self.filePath = filePath;

}

exports.SizeCache = SizeCache;

//-----------------------------------------------------------------------------
// KeyValStore
//-----------------------------------------------------------------------------

function KeyValStore(cacheName, saveInterval, strictCheckValues) {

  if (typeof saveInterval !== 'number') {
    saveInterval = DEF_INTERVAL;
  }

  var self = this;
  var data = {};
  var tout = null;
  var path = null;

  function load(cachPath, cacheFile) {
    if (cacheFile) {
      cachPath = np.join(cachPath, cacheFile);
    }
    var start = Date.now();
    kvdb.loadJSON(cachPath, function(err, kvs) {
      path = cachPath;
      if (err) return;
      for (var key in data) {
        kvs[key] = data[key];
      }
      data = kvs;
      log.debug('Cache %s loaded %d items in: %s sec.', cacheName, count(), fmt.millisPeriod(start));
    });
  }

  function count() {
    var cnt = 0;
    for (var key in data) {
      cnt++;
    }
    return cnt;
  }

  function save(cb) {
    resetTimeout();
    if (path) {
//      log.debug('Saving cache %s', cacheName);
      kvdb.saveJSON(path, data, function(err) {
        if (cb) cb(err)
      });
    }
    else {
      if (cb) cb();
    }
  }

  function resetTimeout() {
    if (tout) {
      clearTimeout(tout);
      tout = null;
    }
  }

  function scheduleSave() {
    resetTimeout();
    tout = setTimeout(save, saveInterval);
  }

  function stat(cb) {

    var cnt  = count();

    if (cnt === 0 || !path) {
      return cb(null, {count: 0, size: 0})
    }

    fop.fileSize(path, function(err, size) {
      if (err) return cb(err);
      cb(null, {count: cnt, size: size})
    });

  }

  function erase(cb) {
    var cnt = count();
    if (cnt === 0) {
      return cb();
    }
    data = {};
    save(cb);
  }

  function get(key) {
    return data[key];
  }

  function set(key, val) {
    if (strictCheckValues) {
      if (data[key] !== val) {
        data[key] = val;
        scheduleSave();
      }
    }
    else {
      data[key] = val;
      scheduleSave();
    }
  }

  function has(key) {
    return (data[key] !== undefined);
  }

  function del(key) {
    delete data[key];
    scheduleSave();
  }

  function findParentKeyForVal(key, val) {
    var par = data[key];
    if (par === undefined) return null;
    if (par === val) return key;
    return findParentKeyForVal(par, val);
  }

  self.get = get;
  self.set = set;
  self.has = has;
  self.del = del;

  self.load  = load;
  self.save  = save;
  self.stat  = stat;
  self.erase = erase;

  self.path = function() { return path };

  self.findParentKeyForVal = findParentKeyForVal;

}

exports.KeyValStore = KeyValStore;

//-----------------------------------------------------------------------------
// JsonFileCache
//-----------------------------------------------------------------------------

function JsonFileCache(cacheName, ext, minlen) {

  var self = this;
  var slot = new SlotMap();
  var path = null;
  var elen = ext.length;

  function load(cachePath) {
    path = cachePath;
    var start = Date.now();
    var count = 0;
    fs.readdir(cachePath, function(err, fnames) {
      if (err) {
        log.error('Cache %s error reading dir: %s', cacheName, cachePath);
        return;
      }
      async.eachLimit(fnames, PARALLEL_LIMIT,
        function(name, acb) {
          var nlen = name.length;
          if (name.charAt(0) === '.' || nlen <  minlen || name.substring(nlen - elen) !== ext) {
            return acb();
          }
          fs.readFile(np.join(cachePath, name),
            function(err, buf) {
              if (err) return acb();
              var key = name.substring(0, nlen - elen);
              slot.set(key, buf);
              count++;
              acb();
            }
          );
        },
        function() {
          log.debug('Cache %s loaded %d items in: %s sec.', cacheName, count, fmt.millisPeriod(start));
        }
      );
    });
  }

  function mkbuf(val) {
    try {
      return new Buffer(JSON.stringify(val));
    }
    catch (e) {
      log.error('Cache %s error mkbuf: %s', cacheName, e);
      return null;
    }
  }

  function save(key, buf) {
    var file = filePath(key);
    fs.writeFile(file, buf, function(err) {
      if (err) {
        log.error('Cache %s write error: %s (%s)', cacheName, file, err);
      }
    });
  }

  function stat(cb) {
    var count = 0;
    var size  = 0;
    slot.each(function(key, buf) {
      count++;
      size += buf.length;
    });
    cb(null, {count: count, size: size});
  }

  function erase(cb) {
    slot.clear();
    fop.deleteFiles(path, cb);
  }

  function filePath(key) {
    return np.join(path,  key + ext);
  }

  function get(key) {
    var val = slot.get(key);
    if (val === undefined) {
      return val;
    }
    try {
      return JSON.parse(val.toString());
    }
    catch (e) {
      log.error('Cache %s error parsing buffer: %s', e);
      slot.del(key);
      return undefined;
    }
  }

  function set(key, val) {
    var buf = mkbuf(val);
    if (buf) {
      slot.set(key, buf);
      setImmediate(save, key, buf);
    }
  }

  function has(key) {
    return slot.has(key);
  }

  function del(key) {
    slot.del(key);
    if (path) {
      fs.unlink(filePath(key), noop);
    }
  }

  function upd(key, val) {
    var buf = mkbuf(val);
    if (buf) slot.set(key, buf);
  }

  // Public Interface

  self.get = get;
  self.set = set;
  self.has = has;
  self.del = del;
  self.upd = upd;

  self.load  = load;
  self.stat  = stat;
  self.erase = erase;

  self.path = function() { return path };

  self.filePath = filePath;

}

exports.JsonFileCache = JsonFileCache;