/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

'use strict';

var util = require('util');
var upnp = require('./upnp');
var xml = require('../xml');
var cfg = require('../cfg');
var rti = require('../rti');
var fmt = require('../fmt');

exports.upnp = upnp;

var notNull = rti.notNull;

//-----------------------------------------------------------------------------

var ns = {
  av:       "urn:schemas-upnp-org:av:av",
  dc:       "http://purl.org/dc/elements/1.1/",
  upnp:     "urn:schemas-upnp-org:metadata-1-0/upnp/",
  didl:     "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
  dlna:     "urn:schemas-dlna-org:metadata-1-0/",
  wmpns:    "urn:schemas-microsoft-com:WMPNSS-1-0/",
  upnp_dev: "urn:schemas-upnp-org:device-1-0",
  dlna_dev: "urn:schemas-dlna-org:device-1-0",
  pnpx:     "http://schemas.microsoft.com/windows/pnpx/2005/11"
};

//-----------------------------------------------------------------------------

var id = {
  id: 'id',
  parentID: 'parentID',
  title: 'title',
  mediaClass: 'mediaClass',
  writeStatus: 'writeStatus',
  creator: 'creator',
  restricted: 'restricted',
  upnpClass: 'upnpClass',
  date: 'date',
  album: 'album',
  author: 'author',
  genre: 'genre',
  publisher: 'publisher',
  albumArtURI: 'albumArtURI',
  // Container properties
  childCount: 'childCount',
  searchable: 'searchable',
  hideCount: 'hideCount',
  multiItem: 'multiItem', // for sorting
  // Resource properties
  res:        'res', // resource array
  url:        'url',
  profile:    'profile',
  resolution: 'resolution',
  colorDepth: 'colorDepth',
  conversion: 'conversion',
  operation:  'operation'
};

exports.id = id;

//-----------------------------------------------------------------------------
// Base Item
//-----------------------------------------------------------------------------

function BaseItem(uid, pid, title) {
  this[id.id] = uid;
  this[id.parentID] = pid;
  this[id.title] = title;
  this[id.restricted] = true;
  this[id.res] = [];
}

BaseItem.prototype.getDate = function() {
  return this[id.date] || 0;
};

BaseItem.prototype.setDate = function(date) {
  this[id.date] = date;
};

BaseItem.prototype.getTitle = function() {
  return this[id.title];
};

BaseItem.prototype.setTitle = function(title) {
  this[id.title] = title;
};

BaseItem.prototype.setMediaClass = function(mediaClass) {
  this[id.mediaClass] = mediaClass;
};

BaseItem.prototype.setUpnpClass = function(upnpClass) {
  this[id.upnpClass] = upnpClass;
};

BaseItem.prototype.addResource = function(res) {
  this[id.res].push(res);
};

BaseItem.prototype.fillXobj = function(xobj) {
  xobj['dc:title'] = this[id.title];
  var val = this[id.upnpClass];
  if (notNull(val)) xobj['upnp:class'] = val;
  val = this[id.date];
  if (notNull(val)) xobj['dc:date'] = dcDate(val);
  val = this[id.mediaClass];
  if (notNull(val)) xobj['av:mediaClass'] = val;
  val = this[id.res];
  if (rti.isArray(val) && val.length > 0) {
    var ary = [];
    for (var i = 0; i < val.length; i++) {
      ary.push(val[i].toXJSON());
    }
    xobj['#list'] = ary;
  }
};

//-----------------------------------------------------------------------------
// Container
//-----------------------------------------------------------------------------

function Container(uid, pid, title, count) {
  BaseItem.call(this, uid, pid, title);
  this[id.upnpClass] = upnp.STORAGE_FOLDER_CONTAINER;
  this[id.childCount] = count || 0;
  this[id.searchable] = false;
  this[id.hideCount]  = false;
  this[id.multiItem]  = false;
}

util.inherits(Container, BaseItem);

Container.prototype.toXJSON = function() {
  var self = this;
  var xobj = {
    container: {
      '@': {
        id:         self[id.id],
        parentID:   self[id.parentID],
        childCount: self[id.childCount],
        searchable: self[id.searchable],
        restricted: self[id.restricted]
      }
    }
  };
  self.fillXobj(xobj.container);
  if (cfg.childCount() && !self[id.hideCount]) {
    xobj.container['dc:title'] += ' [' + fmt.thousandSeparator(self[id.childCount]) + ']';
  }
  return xobj;
};

Container.prototype.setChildCount = function(count) {
  this[id.childCount] = count;
};

Container.prototype.setHideCount = function(hide) {
  this[id.hideCount] = hide;
};

Container.prototype.addChildCount = function(count) {
  this[id.childCount] += count;
};

Container.prototype.isMultiItem = function() {
  return this[id.multiItem];
};

Container.prototype.setMultiItem = function(multi) {
  this[id.multiItem] = multi;
};

exports.Container = Container;

//-----------------------------------------------------------------------------
// Item
//-----------------------------------------------------------------------------

function Item(uid, pid, title) {
  BaseItem.call(this, uid, pid, title);
}

util.inherits(Item, BaseItem);

Item.prototype.toXJSON = function() {
  var self = this;
  var xobj = {
    item: {
      '@': {
        id:         self[id.id],
        parentID:   self[id.parentID],
        restricted: self[id.restricted]
      }
    }
  };
  self.fillXobj(xobj.item);
  return xobj;
};

exports.Item = Item;

exports.imageItem = function(uid, pid, title) {
  var obj = new Item(uid, pid, title);
  obj.setUpnpClass(upnp.IMAGE_ITEM_PHOTO);
  obj.setMediaClass(upnp.AV_PHOTO);
  return obj;
};

//-----------------------------------------------------------------------------
// Resource
//-----------------------------------------------------------------------------

function Res(url, profile, conversion, operation) {
  this[id.url] = url;
  this[id.profile] = profile;
  this[id.conversion] = conversion;
  this[id.operation] = operation;
}

Res.prototype.setProfile = function(profile) {
  this[id.profile] = profile;
};

Res.prototype.setConversion = function(conversion) {
  this[id.conversion] = conversion;
};

Res.prototype.setOperation = function(operation) {
  this[id.operation] = operation;
};

Res.prototype.setResolution = function(width, height) {
  this[id.resolution] = width + 'x' + height;
};

Res.prototype.setColorDepth = function(colorDepth) {
  this[id.colorDepth] = colorDepth;
};

Res.prototype.toXJSON = function() {
  var self = this;
  var xobj = {
    res: {
      '@': {},
      '#text': self[id.url]
    }
  };
  var attr = xobj.res['@'];
  var aval = self[id.profile];
  if (notNull(aval)) {
    attr['protocolInfo'] = aval.toString(false, self[id.conversion], self[id.operation]);
  }
  aval = self[id.resolution];
  if (notNull(aval)) {
    attr[id.resolution] = aval;
  }
  aval = self[id.colorDepth];
  if (notNull(aval)) {
    attr[id.colorDepth] = aval;
  }
  return xobj;
};

exports.Res = Res;

//-----------------------------------------------------------------------------

function didlXML(items) {

  var len = items.length;

  for (var i = 0; i < len; i++) {
    items[i] = items[i].toXJSON();
  }

  var doc = {
    'DIDL-Lite': {
      '@': {
        'xmlns': ns.didl,
        'xmlns:dc': ns.dc,
        'xmlns:av': ns.av,
        'xmlns:upnp': ns.upnp,
        'xmlns:dlna': ns.dlna
      },
      '#list': items
    }
  };

  return xml.toXML(doc, false);

}

exports.didlXML = didlXML;

//-----------------------------------------------------------------------------

function dcDate(d) {

  if (typeof d === 'number') {
    d = new Date(d);
  }

  function pad(number) {
    if (number < 10) {
      return '0' + number;
    }
    return number;
  }

  return d.getUTCFullYear() +
    '-' + pad(d.getUTCMonth() + 1) +
    '-' + pad(d.getUTCDate()) +
    'T' + pad(d.getUTCHours()) +
    ':' + pad(d.getUTCMinutes()) +
    ':' + pad(d.getUTCSeconds());

}