/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

'use strict';

var express = require('express');

var evt = require('../evt');
var acl = require('../acl');
var log = require('../log');
var xml = require('../xml');
var rti = require('../rti');
var cfg = require('../cfg');
var htp = require('../htp');

var upnp = require('./upnp');
var Notify = require('./notify').Notify;

var ROUTER_OPTS = { caseSensitive: true, strict: false };

var id = rti.did;

function RootDevice(name) {

  var self = this;

  var rout;
  var path = '/' + name;
  var info = path + '_info.xml';
  var icon = '/icon';
  var shdr;
  var sary = []; // services
  var uary = [upnp.ROOT_DEVICE]; // USN
  var notify;

  self.iconPath = path + icon;

  self.addService = function(svc) {
    sary.push(svc);
  };

  self.path = path;

  function onGetInfo(req, res) {
    var doc = {
      root: {
        '@': {
          'xmlns':      'urn:schemas-upnp-org:device-1-0',
          'xmlns:dlna': 'urn:schemas-dlna-org:device-1-0',
          'xmlns:pnpx': 'http://schemas.microsoft.com/windows/pnpx/2005/11'
        },
        specVersion: {
          major: 1,
          minor: 0
        },
        device: {
          deviceType:       self[id.deviceType],
          friendlyName:     self[id.friendlyName],
          manufacturer:     self[id.manufacturer],
          manufacturerURL:  self[id.manufacturerURL],
          modelName:        self[id.modelName],
          modelDescription: self[id.modelDescription],
          modelNumber:      self[id.modelNumber],
          modelURL:         self[id.modelURL],
          UDN:              self[id.UDN],
          presentationURL:  null,
          iconList: [],
          serviceList: []
        }
      }
    };

    if (self.icon) {
      doc.root.device.iconList = self.icon.enumIcons(self.iconPath);
    }

    var serviceList = doc.root.device.serviceList;

    for (var i = 0; i < sary.length; i++) {
      serviceList.push(sary[i].description(path));
    }

    if (self.onDocReady) self.onDocReady(doc.root);

    evt.postDlnaInfoReady(self, doc, req);

    htp.setXmlMime(res);

    res.send(xml.toXML(doc));

  }

  function httpLogger(req, res, next) {
    if (log.canHttp()) {
      log.http('%s %s from %s', req.method, req.originalUrl, req.ip);
    }
    next();
  }

  function serverHeader(req, res, next) {
    res.set('Server', shdr);
    next();
  }

  function onGetIcon(req, res) {
    self.icon.serveIcon(req, req.params.name, res);
  }

  // Event hooks

  function onAppReady(app) {

    rout = express.Router(ROUTER_OPTS);

    app.use(path, rout);

    rout.use(serverHeader);
    rout.use(httpLogger);
    rout.use(acl.dlnaAuth);

    app.use(info, serverHeader);
    app.use(info, httpLogger);
    app.use(info, acl.dlnaAuth);
    app.get(info, onGetInfo);


    if (self.icon) {
      rout.get(icon + '/:name', onGetIcon);
    }

    uary.push(self[id.UDN]);
    uary.push(self[id.deviceType]);

    for (var i = 0; i < sary.length; i++) {
      var svc = sary[i];
      uary.push(svc[id.serviceType]);
      svc.initRoute(rout);
    }

  }

  evt.onHttpAppReady(onAppReady);

  function onConfig() {
    shdr = cfg.serverHeader();
    evt.postRootDeviceInit(self);
  }

  evt.onConfigReady(onConfig);

  function onHttpReady() {
    notify = new Notify(self[id.UDN], info, cfg.httpPort(), uary);
    setImmediate(notify.start);
    evt.addExitProc(notify.stop);
  }

  evt.onHttpSrvReady(onHttpReady);

  self.notifyAll = function(cb) {
    if (notify) {
      notify.notifyAll(cb);
    }
    else {
      if (cb) cb();
    }
  };

  self.notifyByeBye = function(cb) {
    if (notify) {
      notify.notifyByeBye(cb);
    }
    else {
      if (cb) cb();
    }
  };

}

exports.RootDevice = RootDevice;