/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

'use strict';

var events = require('events');
var util   = require('util');
var dgram  = require('dgram');
var crypto = require('crypto');
var async  = require('async');

var upnp = require('./upnp');

var log = require('../log');
var evt = require('../evt');
var cfg = require('../cfg');

var SKIP_REPEATED = 2000; // 2 secs

var UPNP_HOST = upnp.HOST;
var UPNP_PORT = upnp.PORT;

var SSDP_ALL    = 'ssdp:all';
var SSDP_ALIVE  = 'ssdp:alive';
var SSDP_BYEBYE = 'ssdp:byebye';
var SSDP_UPDATE = 'ssdp:update';

exports.SSDP_ALL = SSDP_ALL;

exports.ALIVE  = SSDP_ALIVE;
exports.BYEBYE = SSDP_BYEBYE;
exports.UPDATE = SSDP_UPDATE;

var EOL = '\r\n';

var T_OTHER  = 0;
var T_NOTIFY = 1;
var T_SEARCH = 2;
var T_HTTP11 = 3;

var T_STRING = {
  'NOTIFY':   T_NOTIFY,
  'M-SEARCH': T_SEARCH,
  'HTTP/1.1': T_HTTP11
};

//-----------------------------------------------------------------------------
// Skip Addr
//-----------------------------------------------------------------------------

var skipmap = {};

function addSkip(addr) {
  var key = addrToString(addr);
  if (key) {
    skipmap[key] = true;
  }
}

function delSkip(addr) {
  var key = addrToString(addr);
  if (key) {
    delete skipmap[key];
  }
}

function hasSkip(addr) {
  var key = addrToString(addr);
  return (key && skipmap[key] === true);
}

function addrToString(addr) {
  if (addr && addr.address) {
    return addr.address + ':' + addr.port;
  }
  return undefined;
}

function normalizeFrom(from) {
  from.addr = from.address;
  delete from.address;
  delete from.family;
}

//-----------------------------------------------------------------------------
// UDP Cache
//-----------------------------------------------------------------------------

function UDPCache(expire) {

  var self = this;
  var cmap = {};

  function check(data, from, when) {

    var md5 = crypto.createHash('md5');
    md5.update(data);
    var hash = md5.digest('hex');

    var addr = from.address;
    var amap = cmap[addr];

    if (!amap) {
      amap = {};
      cmap[addr] = amap;
    }

    var last = amap[hash] || 0;

    amap[hash] = when;

    return ((when - last) > expire);

  }

  self.check = check;

}

//-----------------------------------------------------------------------------
// Request Class
//-----------------------------------------------------------------------------

function Message(face, from, type, hmap, text, when) {

  var self = this;

  self.face = face;
  self.from = from;
  self.type = type;
  self.text = text;
  self.when = when;
  self.hmap = hmap;

  function h(key) {
    return hmap[key.toUpperCase()];
  }

  self.h = h;

  self.isKnown   = function() { return (type > T_OTHER) };
  self.isNotify  = function() { return (type === T_NOTIFY) };
  self.isSearch  = function() { return (type === T_SEARCH) };
  self.isHttp11  = function() { return (type === T_HTTP11) };

  self.isAlive   = function() { return (h('NTS') === SSDP_ALIVE) };
  self.isByeBye  = function() { return (h('NTS') === SSDP_BYEBYE) };

  self.nt  = h('NT');
  self.nts = h('NTS');

  self.uuid = null;
  self.usn  = null;

  if (type === T_NOTIFY) {
    self.usn = h('NT');
  }
  else if (type === T_HTTP11) {
    self.usn = h('ST');
  }
  else if (type === T_SEARCH) {
    self.st = h('ST');
    self.mx = parseInt(h('MX'));
    if (isNaN(self.mx)) self.mx = 1;
  }

  self.href = h('LOCATION');

  var uusn = h('USN');

  if (uusn) {
    var ary = uusn.split(':');
    if (ary[0] === 'uuid' && ary.length > 1) {
      self.uuid = ary[1];
    }
  }

}

//-----------------------------------------------------------------------------
// Message Parser
//-----------------------------------------------------------------------------

function parseMessage(face, from, data, when) {
  var hmap = {};
  var text = data.toString('ascii');
  text = text.trim();
  var lary = text.split(EOL);
  var line = lary[0].split(' ')[0];
  var type = T_STRING[line];
  if (type === undefined) type = T_OTHER;
  for (var i = 1; i < lary.length; i++) {
    line = lary[i].trim();
    var llen = line.length;
    if (llen === 0) continue;
    var pos = line.indexOf(':');
    if (pos < 1) continue;
    var key = line.substring(0, pos).trim();
    var val = line.substring(pos + 1).trim();
    if (key.length > 0) {
      hmap[key.toUpperCase()] = val;
    }
  }
  return new Message(face, from, type, hmap, text, when);
}

//-----------------------------------------------------------------------------
// Listener Class
//-----------------------------------------------------------------------------

function Listener(face, wall) {

  var self = this;

  events.EventEmitter.call(self);

  var sock = null;
  var isok = false;
  var cache = new UDPCache(SKIP_REPEATED);

  function closeSocket() {
    if (sock) {
      try { sock.close() } catch (e) {}
      sock = null;
    }
  }

  // Incoming Events

  function onMessage(data, from) {

    try {

      if (hasSkip(from)) return;

      normalizeFrom(from);

      if (wall && !wall(face, from)) {
        return;
      }

      var when = Date.now();

      if (!cache.check(data, from, when)) {
        return;
      }

      var msg = parseMessage(face, from, data, when);

      setImmediate(evt.postSsdpMessage, msg);

    }
    catch (e) {
      log.error('SSDP onmessage at %s error: %s', face.addr, e);
    }

  }

  function onError(err) {
    isok = false;
    closeSocket();
    log.error('SSDP socket error: %s (%s)', face.addr, err);
  }

  function onClose() {
    isok = false;
    log.ssdp('Socket closed: %s', face.addr);
  }

  function onListening() {
    sock.addMembership(UPNP_HOST, face.addr);
    isok = true;
    log.ssdp('Listening: %s', face.addr);
  }

  // Init

  function init() {

    sock = dgram.createSocket('udp4');

    sock.on('message', onMessage);
    sock.on('error', onError);
    sock.on('close', onClose);
    sock.on('listening', onListening);

    sock.bind(UPNP_PORT);

  }

  // Done

  function done() {
    if (isok) {
      isok = false;
      closeSocket();
    }
  }

  //-----------------------------------------------------------------------------
  // Public Interface
  //-----------------------------------------------------------------------------

  self.init = init;
  self.done = done;
  self.addr = function() { return face.addr };
  self.name = function() { return face.name };
  self.isok = function() { return isok };

}

util.inherits(Listener, events.EventEmitter);

exports.Listener = Listener;

//-----------------------------------------------------------------------------
// Search
//-----------------------------------------------------------------------------

exports.search = function(face, wall, st, mx, ttl, cb) {

  if (typeof mx !== 'number' || mx < 1 || mx > 5) mx = 3;
  if (typeof st !== 'string' || st.length == 0) st = 'ssdp:all';
  if (typeof ttl !== 'number' || ttl < 1 || ttl > 255) ttl = 4;

  var header = [
    'M-SEARCH * HTTP/1.1',
    'HOST: 239.255.255.250:1900',
    'MAN: "ssdp:discover"',
    'MX: ' + mx,
    'ST: ' + st,
    EOL
  ];

  var timeout = (mx + 2) * 1000;

  var packet = new Buffer(header.join(EOL), 'ascii');

  var sock = dgram.createSocket('udp4');
  var addr;
  var cache = new UDPCache(timeout);

  sock.on('message', onMessage);
  sock.on('error', onError);
  sock.on('listening', onListening);

  sock.bind(0, face.addr);

  // Search Response

  function onMessage(data, from) {
    try {
      normalizeFrom(from);
      if (wall && !wall(face, from)) {
        return;
      }
      var when = Date.now();
      if (!cache.check(data, from, when)) {
        return;
      }
      var msg = parseMessage(face, from, data, when);
      setImmediate(evt.postSsdpMessage, msg);
    }
    catch (e) {
      log.error('SSDP search message at %s error: %s', face.addr, e);
    }

  }

  function onError(err) {
    closeSocket();
    log.error('SSDP search socket error: %s (%s)', face.addr, err);
  }

  function onListening() {
    try {
      addr = sock.address();
      addSkip(addr);
      sock.setMulticastTTL(ttl);
      sock.setMulticastLoopback(false);
      setImmediate(sendPacket);
      setTimeout(closeSocket, timeout);
    } catch (e) {
      log.error('SSDP search listening at %s error: %s', face.addr, e);
      closeSocket();
    }
  }

  var sendCount = 0;

  function sendPacket() {
    try {
      sock.send(packet, 0, packet.length, UPNP_PORT, UPNP_HOST, function(err, bytes) {
        if (err) {
          log.error('SSDP search send error on: %s (%s)', face.addr, err);
          closeSocket();
        }
        else {
          if (sendCount++ < 3) {
            setTimeout(sendPacket, 100);
          }
        }
      });
    } catch (e) {
      log.error('SSDP search listening at %s error: %s', face.addr, e);
      closeSocket();
    }
  }

  function closeSocket() {
    if (sock) {
      delSkip(addr);
      try { sock.close() } catch (e) {}
      sock = null;
    }
    if (cb) {
      cb();
      cb = null;
    }
  }

};

//-----------------------------------------------------------------------------
// Notify
//-----------------------------------------------------------------------------

function makePacket(method, header) {
  var text = method + EOL;
  for (var key in header) {
    text += key + ': ' + header[key] + EOL;
  }
  text += EOL;
  return new Buffer(text, 'ascii');
}

exports.notify = function(face, uuid, href, uary, verb, cb) {

  var ttl = cfg.udpTTL();

  if (typeof ttl !== 'number' || ttl < 1 || ttl > 255) ttl = 4;

  var nts = verb;

  var method = 'NOTIFY * HTTP/1.1';

  var header = {
    'HOST': '239.255.255.250:1900',
    'CACHE-CONTROL': 'max-age=1800',
    'LOCATION': href,
    'SERVER': cfg.serverHeader(),
    'NTS': nts,
    'NT' : null,
    'USN': null
  };

  var sock = dgram.createSocket('udp4');
  var addr;

  sock.on('error', onError);
  sock.on('listening', onListening);

  sock.bind(0, face.addr);

  function onError(err) {
    closeSocket();
    log.error('SSDP notify socket error: %s (%s)', face.addr, err);
  }

  function onListening() {
    try {
      addr = sock.address();
      addSkip(addr);
      sock.setMulticastTTL(ttl);
      sock.setMulticastLoopback(false);
      async.eachSeries(uary, sendUSN, closeSocket);
    } catch (e) {
      log.error('SSDP notify listening at %s error: %s', face.addr, e);
      closeSocket();
    }
  }

  function sendUSN(usn, scb) {
    header['NT']  = usn;
    header['USN'] = uuid;
    if (usn !== uuid) {
      header['USN'] += '::' + usn;
    }
    var packet = makePacket(method, header);
    if (log.canSsdp()) {
//      log.ssdp('SSDP notify to: %s\n%s', face.addr, log.border(packet.toString().trim()));
      log.ssdp('Sending %s to %s %s', nts, face.addr, usn);
    }
    sock.send(packet, 0, packet.length, UPNP_PORT, UPNP_HOST, function(err) {
      if (err) return scb(err);
      sock.send(packet, 0, packet.length, UPNP_PORT, UPNP_HOST, scb);
    });
  }

  function closeSocket() {
    if (sock) {
      delSkip(addr);
      try { sock.close() } catch (e) {}
      sock = null;
    }
    if (cb) {
      cb();
      cb = null;
    }
  }

};

//-----------------------------------------------------------------------------
// Search Response
//-----------------------------------------------------------------------------

//HTTP/1.1 200 OK
//CACHE-CONTROL: max-age=1800
//DATE: Sat, 02 Jan 2010 04:19:22 GMT
//EXT:
//LOCATION: http://10.0.0.34:2870/dmr.xml
//SERVER: IPI/1.0 UPnP/1.0 DLNADOC/1.50
//ST: upnp:rootdevice
//USN: uuid:140479c0-58f3-1cef-84bf-3cbdd862c7a1::upnp:rootdevice

exports.searchResponse = function(msg, uuid, href, uary, cb) {

  var method = 'HTTP/1.1 200 OK';

  var header = {
    'CACHE-CONTROL': 'max-age=1200',
    'EXT': '',
    'DATE': new Date().toUTCString(),
    'LOCATION': href,
    'SERVER': cfg.serverHeader(),
    'ST' : null,
    'USN': null
  };

  var port = msg.from.port;
  var addr = msg.from.addr;

  var sock = dgram.createSocket('udp4');

  sock.on('error', onError);
  sock.on('listening', onListening);

  sock.bind();

  function onError(err) {
    log.error('SSDP search response socket error: %s', err);
    closeSocket();
  }

  function onListening() {
    sock.setTTL(cfg.udpTTL());
    async.eachSeries(uary, sendUSN, closeSocket);
  }

  function sendUSN(usn, scb) {
    header['ST']  = usn;
    header['USN'] = uuid;
    if (usn !== uuid) {
      header['USN'] += '::' + usn;
    }
    var packet = makePacket(method, header);
    sock.send(packet, 0, packet.length, port, addr, function(err) {
      if (err) {
        log.error('SSDP search response error: %s', err);
        return scb(err);
      }
      scb();
    });
    if (log.canSsdp()) {
//      log.ssdp('Search response to %s:%d\n%s', addr, port, log.border(packet.toString().trim()));
      log.ssdp('Search response to %s:%d %s', addr, port, usn);
    }
  }

  function closeSocket() {
    if (sock) {
      try { sock.close() } catch (e) {}
      sock = null;
    }
    if (cb) {
      cb();
      cb = null;
    }
  }

};
