/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

'use strict';

var async = require('async');
var EventEimmiter = require('events').EventEmitter;

var EXIT            = "exit";
var HTTP_SRV_READY  = "httpSrvReady";
var HTTP_APP_READY  = "httpAppReady";
var AGENT_READY     = "agentReady";
var CONFIG_READY    = "configReady";
var PID_FILE_READY  = "pidFileReady";
var INFO_REQUEST    = "info_request";
var DLNA_INFO_READY = "dlnaInfoReady";
var NET_FACE_ON     = "netFaceOn";
var NET_FACE_OFF    = "netFaceOff";
var CONFIG_CHANGE   = "configChange";

var SSDP_SEARCH_TARGET = "ssdpSearchTarget";
var SSDP_MESSAGE = "ssdpMessage";

var DEV_CREATE = 'devCreate';
var DEV_REMOVE = 'devRemove';
var DEV_UPDATE = 'devUpdate';
var ROOT_DEVICE_INIT = 'rootDeviceInit';
var DLNA_SOURCE = 'dlnaSource';
var GET_SERVERS = 'getServers';

var BROWSE_REQUEST = 'browseRequest';
var BEFORE_IMAGE_STREAM = 'beforeImageStream';
var SOAP_REQUEST = 'soapRequest';
var START_STREAM = 'startStream';
var STOP_STREAM  = 'stopStream';

var emitter = new EventEimmiter();

emitter.setMaxListeners(500); // Yes, we love listeners!

var exiting = false;

var exitProcs = [];

function logError(name, err) {
  try {
    var log = require('./log');
    log.error('Event "%s" error: %s', name, err);
  } catch (e) {
    // ignore
  }
}

exports.exiting = function() { return exiting; };

exports.quit = function(code) {
  if (!exiting) {
    code = code || 0;
    exiting = true;
    emitter.emit(EXIT);
    async.parallel(exitProcs, function() {
      require('./log').end(function() {
        process.exit(code);
      });
    });
  }
};

exports.addExitProc = function(proc) {
  exitProcs.push(proc);
};

exports.async = function(proc) {
  setImmediate(proc);
};

exports.throwGlobal = function(err) {
  setImmediate(function(e) {throw  e}, err);
};

exports.postHttpSrvReady = function(address) {
  emitter.emit(HTTP_SRV_READY, address);
};

exports.postHttpAppReady = function(app) {
  emitter.emit(HTTP_APP_READY, app);
};

exports.postAgentReady = function() {
  emitter.emit(AGENT_READY);
};

exports.postConfigReady = function() {
  emitter.emit(CONFIG_READY);
};

exports.postConfigChange = function() {
  emitter.emit(CONFIG_CHANGE);
};

exports.postPidFileReady = function() {
  emitter.emit(PID_FILE_READY);
};

exports.postInfoRequest = function(req) {
  emitter.emit(INFO_REQUEST, req);
};

exports.postDlnaInfoReady = function(dev, doc, req) {
  emitter.emit(DLNA_INFO_READY, dev, doc, req);
};

exports.postNetFaceOn = function(face) {
  emitter.emit(NET_FACE_ON, face);
};

exports.postNetFaceOff = function(face) {
  emitter.emit(NET_FACE_OFF, face);
};

exports.postSsdpSearchTarget = function(targetsArray) {
  emitter.emit(SSDP_SEARCH_TARGET, targetsArray);
};

exports.postSsdpMessage = function(msg) {
  emitter.emit(SSDP_MESSAGE, msg);
};

exports.postDevCreate = function(dev, doc) {
  emitter.emit(DEV_CREATE, dev, doc);
};

exports.postDevRemove = function(dev) {
  emitter.emit(DEV_REMOVE, dev);
};

exports.postDevUpdate = function(dev) {
  emitter.emit(DEV_UPDATE, dev);
};

exports.postRootDeviceInit = function(dev) {
  emitter.emit(ROOT_DEVICE_INIT, dev);
};

exports.postDlnaSource = function(sourceArray) {
  emitter.emit(DLNA_SOURCE, sourceArray);
};

exports.postBrowseRequest = function(brq) {
  emitter.emit(BROWSE_REQUEST, brq);
};

exports.postGetServers = function(ary) {
  emitter.emit(GET_SERVERS, ary);
};

exports.postBeforeImageStream = function(srq, imagePath) {
  emitter.emit(BEFORE_IMAGE_STREAM, srq, imagePath);
};

exports.postSoapRequest = function(req, act, arg) {
  emitter.emit(SOAP_REQUEST, req, act, arg);
};

exports.postStartStream = function() {
  emitter.emit(START_STREAM);
};

exports.postStopStream = function() {
  emitter.emit(STOP_STREAM);
};

//-----------------------------------------------------------------------------

exports.onExit = function(listener) {
  emitter.addListener(EXIT, listener);
};

exports.onHttpSrvReady = function(listener) {
  emitter.addListener(HTTP_SRV_READY, listener);
};

exports.onHttpAppReady = function(listener) {
  emitter.addListener(HTTP_APP_READY, listener);
};

exports.onAgentReady = function(listener) {
  emitter.addListener(AGENT_READY, listener);
};

exports.onConfigReady = function(listener) {
  emitter.addListener(CONFIG_READY, listener);
};

exports.onConfigChange = function(listener) {
  emitter.addListener(CONFIG_CHANGE, listener);
};

exports.onConfigReadyChange = function(listener) {
  emitter.addListener(CONFIG_READY, listener);
  emitter.addListener(CONFIG_CHANGE, listener);
};

exports.onPidFileReady = function(listener) {
  emitter.addListener(PID_FILE_READY, listener);
};

exports.onInfoRequest = function(listener) {
  emitter.addListener(INFO_REQUEST, listener);
};

exports.onDlnaInfoReady = function(listener) {
  emitter.addListener(DLNA_INFO_READY, listener);
};

exports.onNetFaceOn = function(listener) {
  emitter.addListener(NET_FACE_ON, listener);
};

exports.onNetFaceOff = function(listener) {
  emitter.addListener(NET_FACE_OFF, listener);
};

exports.onSsdpSearchTarget = function(listener) {
  emitter.addListener(SSDP_SEARCH_TARGET, listener);
};

exports.onSsdpMessage = function(listener) {
  emitter.addListener(SSDP_MESSAGE, listener);
};

exports.onDevCreate = function(listener) {
  emitter.addListener(DEV_CREATE, listener);
};

exports.onDevRemove = function(listener) {
  emitter.addListener(DEV_REMOVE, listener);
};

exports.onDevUpdate = function(listener) {
  emitter.addListener(DEV_UPDATE, listener);
};

exports.onRootDeviceInit = function(listener) {
  emitter.addListener(ROOT_DEVICE_INIT, listener);
};

exports.onDlnaSource = function(listener) {
  emitter.addListener(DLNA_SOURCE, listener);
};

exports.onBrowseRequest = function(listener) {
  emitter.addListener(BROWSE_REQUEST, listener);
};

exports.onGetServers = function(listener) {
  emitter.addListener(GET_SERVERS, listener);
};

exports.onBeforeImageStream = function(listener) {
  emitter.addListener(BEFORE_IMAGE_STREAM, listener);
};

exports.onSoapRequest = function(listener) {
  emitter.addListener(SOAP_REQUEST, listener);
};

exports.onStartStream = function(listener) {
  emitter.addListener(START_STREAM, listener);
};

exports.onStopStream = function(listener) {
  emitter.addListener(STOP_STREAM, listener);
};



