/*
 * Copyright (c) 2014 iSedora.com - All Rights Reserved
 * This file is part of iSedora Media Server software
 */

// iSedora iPhoto Plugin

'use strict';

var fs = require('fs');
var np = require('path');
var async = require('async');

var agt = isedora['agt'];
var log = isedora['log'];
var evt = isedora['evt'];
var env = isedora['env'];
var cfg = isedora['cfg'];
var hnd = isedora['hnd'];
var rti = isedora['rti'];
var htp = isedora['htp'];
var fwr = isedora['cls']['fwatcher'];
var mdi = isedora['dlna']['mdi'];
var img = isedora['dlna']['img'];

var didl = isedora['cls']['didl'];
var upnp = didl.upnp;

var Container = didl.Container;

var mtype = rti.mtype;

var PLUGIN_NAME = 'isedora-iphoto';
var HANDLE_NAME = 'iph';

var IPHOTO_FOLDER   = 'iPhoto Library';
var ALBUM_DATA_FILE = 'AlbumData.xml';
var DATE_INTERVAL   = Date.UTC(2001, 0, 1, 0, 0, 0, 0);

var enabled = isEnabled();
var watcher = new fwr.FileWatcher(onChange);

var TOP_ID = 'top';
var EVT_ID = 'events';
var ALB_ID = 'a';

// iPhoto Parsed Data

var idata = {
  master: null,
  albums: null,
  events: [],
  direct: []
};

var icfg = {};

// iPhoto Keys

var pk = {
  masterImageList: 'Master Image List',
  caption:         'Caption',
  imagePath:       'ImagePath',
  originalPath:    'OriginalPath',
  guid:            'GUID',
  mediaType:       'MediaType',
  modDate:         'ModDateAsTimerInterval',
  date:            'DateAsTimerInterval',
  listOfAlbums:    'List of Albums',
  listOfRolls:     'List of Rolls',
  master:          'Master',
  albumName:       'AlbumName',
  photoCount:      'PhotoCount',
  keyList:         'KeyList',
  albumType:       'Album Type',
  rollID:          'RollID',
  rollName:        'RollName',
  rollDate:        'RollDateAsTimerInterval',
  sorted:          '_sorted'
};

// iPhoto Vals

var pv = {
  allPhotosAlbum:   'allPhotosAlbum',
  lastImportAlbum:  'lastImportAlbum',
  emptyEventAlbum:  'emptyEventAlbum',
  lastNMonthsAlbum: 'lastNMonthsAlbum',
  flaggedAlbum:     'flaggedAlbum',
  eventAlbumType:   'Event',
  regularAlbumType: 'Regular'
};

//-----------------------------------------------------------------------------
// iPhoto Browser
//-----------------------------------------------------------------------------

function iphotoBrowser(brq, cb) {
  var uid = brq.uid;
  var act = uid.charAt(0);
  brq.uid = uid.substring(1);
  switch (act) {
    case 't': // Top browse
      return topBrowser(brq, cb);
    case 'e': // Events browse
      return eventsBrowser(brq, cb);
    case 'a': // Album browse
      return albumBrowser(brq, cb);
    default:
      log.warn('iPhoto handler not found for: %s', brq.oid);
  }
  cb();
}

hnd.addBrowseHandler(HANDLE_NAME, iphotoBrowser);

//-----------------------------------------------------------------------------
// Photo Count
//-----------------------------------------------------------------------------

hnd.addPhotoCount(function(brq) {
  return (enabled ? 1 : 0);
});

//-----------------------------------------------------------------------------
// Photo Browser
//-----------------------------------------------------------------------------

function photoBrowser(brq, cb) {
  if (enabled && !brq.meta) {
    var top = topFolder(brq);
    var ix = brq.indexOfMediaClass(upnp.AV_PHOTO);
    if (ix > -1) {
      brq.insert(ix, top);
    }
    else {
      brq.append(top);
    }
  }
  cb();
}

hnd.addPhotoBrowser(photoBrowser);

//-----------------------------------------------------------------------------
// Top Browser
//-----------------------------------------------------------------------------

function topBrowser(brq, cb) {
  if (brq.meta) {
    brq.append(topFolder(brq));
    return cb();
  }
  brq.append(eventsFolder(brq));
  var direct = idata.direct;
  var dirlen = direct.length;
  for (var i = 0; i < dirlen; i++) {
    var alb = albumFolder(brq, direct[i]);
    if (alb) {
      brq.append(alb);
    }
  }
  cb();
}

//-----------------------------------------------------------------------------
// Events Browser
//-----------------------------------------------------------------------------

function eventsBrowser(brq, cb) {
  if (brq.meta) {
    brq.append(eventsFolder(brq));
    return cb();
  }
  var events = idata.events;
  var evtlen = events.length;
  for (var i = 0; i < evtlen; i++) {
    var alb = albumFolder(brq, events[i]);
    if (alb) {
      brq.append(alb);
    }
  }
  cb();
}

//-----------------------------------------------------------------------------
// Album Browser
//-----------------------------------------------------------------------------

function albumBrowser(brq, cb) {
  if (brq.meta) {
    var albfolder = albumFolder(brq, brq.uid);
    if (albfolder) {
      brq.append(albfolder);
    }
    return cb();
  }
  var alb = idata.albums[brq.uid];
  if (!alb) {
    return cb('iPhoto album not found: ' + brq.oid);
  }
  var keys = alb[pk.keyList];
  var klen = keys.length;
  if (klen === 0) return cb();
  if (!alb[pk.sorted]) {
    sortPhotos(idata.master, keys, icfg.photoSort);
    alb[pk.sorted] = true;
  }
  var start = brq.start;
  var count = brq.count;
  brq.range = true;
  brq.total = klen;
  if (start >= klen) return cb();
  var rkeys = [];
  for (var i = start; i < (start + count) && i < klen; i++) {
    rkeys.push(keys[i]);
  }
  var master = idata.master;
  async.map(rkeys,
    function(key, pcb) {
      var photo = master[key];
      if (!photo) {
        log.warn('iPhoto master image not found: %s', key);
        return cb(null, null);
      }
      var fpath = photo[pk.imagePath];
      mdi.fileInfo(fpath, 0, true, function(err, md) {
        if (!rti.isImage(md)) {
          brq.total--;
          return pcb(null, null);
        }
        pcb(null, imageItem(brq, md, photo));
      });
    },
    function(err, pdidl) {
      brq.addall(pdidl);
      cb();
    }
  );
}

//-----------------------------------------------------------------------------
// Folders & Item
//-----------------------------------------------------------------------------

function topFolder(brq) {
  var top = new Container(mkoid(TOP_ID), brq.oid, 'iPhoto', topCount());
  top.setHideCount(true);
  top.setMediaClass(upnp.AV_PHOTO);
  return top;
}

function topCount() {
  var cnt = 0;
  cnt++; // Events
  cnt += idata.direct.length;
  return cnt;
}

function eventsFolder(brq) {
  var evt = new Container(mkoid(EVT_ID), brq.oid, 'Events', idata.events.length);
  evt.setMediaClass(upnp.AV_PHOTO);
  return evt;
}

function albumFolder(brq, guid) {
  var alb = idata.albums[guid];
  if (!alb) {
    log.warn('iPhoto album not found: %s', guid);
    return null;
  }
  var count = alb[pk.photoCount];
  if (!icfg.showEmpty && count === 0) {
    return null;
  }
  var af = new Container(mkoid(ALB_ID + guid), brq.oid, alb[pk.albumName], count);
  af.setMediaClass(upnp.AV_PHOTO);
  return af;
}

function imageItem(brq, md, photo) {
  var item = img.makeImageItem(brq, md);
  item.setTitle(photo[pk.caption]);
  item.setDate(new Date((photo[pk.date] * 1000) + DATE_INTERVAL));
  return item;
}

//-----------------------------------------------------------------------------

function mkoid(name) {
  return hnd.mkoid(HANDLE_NAME, name);
}

//-----------------------------------------------------------------------------

function onConfigChange() {
  enabled = isEnabled();
  if (enabled) {
    var ncfg = cfg.getPluginConfig(PLUGIN_NAME);
    var fpath = ncfg.path;
    if (fpath && watcher.dirname() !== fpath) {
      icfg = ncfg;
      setImmediate(load, fpath);
    }
    else {
      if (icfg.eventSort !== ncfg.eventSort) {
        sortAlbums(idata.albums, idata.events, ncfg.eventSort);
      }
      if (icfg.photoSort !== ncfg.photoSort) {
        var albums = idata.albums;
        for (var key in albums) {
          var alb = albums[key];
          alb[pk.sorted] = false;
        }
      }
      icfg = ncfg;
    }
  }
  else {
    watcher.stop();
  }
}

evt.onConfigChange(onConfigChange);

//-----------------------------------------------------------------------------

function isEnabled() {
  return cfg.pluginEnabled(PLUGIN_NAME);
}

//-----------------------------------------------------------------------------

function onChange(err) {
  if (err) {
    log.error('iPhoto file watch error: %s (%s)', watcher.fpath(), err);
    return;
  }
  log.debug('iPhoto file changed: %s', watcher.fpath());
  load(getIphotoFolder());
}

//-----------------------------------------------------------------------------

function findIphotoFolder() {
  var sysimg = cfg.getSysImageFolders();
  if (Array.isArray(sysimg)) {
    for (var i = 0; i < sysimg.length; i++) {
      var path = np.join(sysimg[i].path, IPHOTO_FOLDER);
      if (fs.existsSync(np.join(path, ALBUM_DATA_FILE))) {
        return path;
      }
    }
  }
  return null;
}

//-----------------------------------------------------------------------------

function getIphotoFolder() {
  return icfg.path;
}

//-----------------------------------------------------------------------------

function load(fpath) {
  watcher.stop();
  fpath = np.join(fpath, ALBUM_DATA_FILE);
  log.debug('Loading iPhoto data: %s', fpath);
  agt.plist(fpath, function(err, rdata) {
    if (err) {
      log.error('Error loading iPhoto data: %s (%s)', fpath, err);
      return;
    }
    try {
      parseData(rdata);
      log.debug('iPhoto data loaded');
      watcher.start(fpath);
    } catch (e) {
      log.error('Error parsing iPhoto data: %s', e);
    }
  });
}

//-----------------------------------------------------------------------------

function parseData(rdata) {
  idata.master = rdata[pk.masterImageList];
  idata.albums = {};
  idata.events = [];
  idata.direct = [];
  // Rolls
  var rolls = rdata[pk.listOfRolls];
  var rolen = rolls.length;
  for (var i = 0; i < rolen; i++) {
    var r = rolls[i];
    var rid = 'r' + r[pk.rollID];
    r[pk.guid] = rid;
    r[pk.albumName] = r[pk.rollName];
    r[pk.date] = r[pk.rollDate];
    r[pk.sorted] = false;
    idata.albums[rid] = r;
    idata.events.push(rid);
  }
  sortAlbums(idata.albums, idata.events, icfg.eventSort);
  // Albums
  var albary = rdata[pk.listOfAlbums];
  var alblen = albary.length;
  for (i = 0; i < alblen; i++) {
    var alb = albary[i];
    var atp = alb[pk.albumType];
    if (atp === pv.eventAlbumType) continue;
    alb[pk.sorted] = false;
    var uid = alb[pk.guid];
    idata.albums[uid] = alb;
    idata.direct.push(uid);
  }
}

//-----------------------------------------------------------------------------

function iPhotoFolderFilter(brq, md, cb) {
  if (enabled && brq.isImage() && rti.isFolder(md) && md[rti.id.fname] === IPHOTO_FOLDER) {
    if (rti.fpath(md) === np.dirname(watcher.fpath())) {
      cb(null, null);
      return true;
    }
  }
}

hnd.addMediaHandler(iPhotoFolderFilter);

//-----------------------------------------------------------------------------
// Sorting
//-----------------------------------------------------------------------------

function sortAlbums(albums, guidary, order) {

  function newest(id1, id2) {
    var a1 = albums[id1];
    var a2 = albums[id2];
    var d1 = a1[pk.date];
    var d2 = a2[pk.date];
    return d2 - d1;
  }

  function oldest(id1, id2) {
    var a1 = albums[id1];
    var a2 = albums[id2];
    var d1 = a1[pk.date];
    var d2 = a2[pk.date];
    return d1 - d2;
  }

  function byname(id1, id2) {
    var a1 = albums[id1];
    var a2 = albums[id2];
    var n1 = a1[pk.albumName];
    var n2 = a2[pk.albumName];
    return n1.toLowerCase().localeCompare(n2.toLowerCase());
  }

  switch (order) {
    case rti.sort.newest:
      guidary.sort(newest);
      break;
    case rti.sort.oldest:
      guidary.sort(oldest);
      break;
    case rti.sort.name:
      guidary.sort(byname);
      break;
  }

}

function sortPhotos(master, keylist, order) {

  function newest(id1, id2) {
    var a1 = master[id1];
    var a2 = master[id2];
    var d1 = a1[pk.date];
    var d2 = a2[pk.date];
    return d2 - d1;
  }

  function oldest(id1, id2) {
    var a1 = master[id1];
    var a2 = master[id2];
    var d1 = a1[pk.date];
    var d2 = a2[pk.date];
    return d1 - d2;
  }

  function byname(id1, id2) {
    var a1 = master[id1];
    var a2 = master[id2];
    var n1 = a1[pk.caption];
    var n2 = a2[pk.caption];
    return n1.toLowerCase().localeCompare(n2.toLowerCase());
  }

  switch (order) {
    case rti.sort.newest:
      keylist.sort(newest);
      break;
    case rti.sort.oldest:
      keylist.sort(oldest);
      break;
    case rti.sort.name:
      keylist.sort(byname);
      break;
  }

}

//-----------------------------------------------------------------------------
// Main
//-----------------------------------------------------------------------------

function main() {
  icfg = cfg.getPluginConfig(PLUGIN_NAME);
  if (!icfg) {
    icfg = {
      path: findIphotoFolder(),
      eventSort: rti.sort.newest,
      photoSort: rti.sort.oldest,
      showEmpty: true
    };
    cfg.setPluginConfig(PLUGIN_NAME, icfg);
  }
  var fpath = getIphotoFolder();
  if (enabled && fpath) {
    load(fpath);
  }
}

setImmediate(main);